/*
 * Antonio Barbalace, SSRG Virginia Tech, 2015
 * 
 * Most of the code derives from the Linux kernel
 * Documentation/vm/page-types.c
 * 
 * compile gcc -c dproc dump_process.c
 * run ./dproc PID 1> PID.dump
 */

#define _LARGEFILE64_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <stdarg.h>
#include <string.h>
#include <getopt.h>
#include <limits.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/fcntl.h>
#include <sys/mount.h>
#include <sys/statfs.h>

#ifndef MAX_PATH
# define MAX_PATH 256
#endif

/*
 * application's flags
 */
#define VMA_READ 0
#define VMA_WRITE 1
#define VMA_EXEC 2
#define VMA_SHARE 3

/*
 * pagemap kernel ABI bits
 */

#define PM_ENTRY_BYTES      sizeof(uint64_t)
#define PM_STATUS_BITS      3
#define PM_STATUS_OFFSET    (64 - PM_STATUS_BITS)
#define PM_STATUS_MASK      (((1LL << PM_STATUS_BITS) - 1) << PM_STATUS_OFFSET)
#define PM_STATUS(nr)       (((nr) << PM_STATUS_OFFSET) & PM_STATUS_MASK)

#define PM_PSHIFT_BITS      6
#define PM_PSHIFT_OFFSET    (PM_STATUS_OFFSET - PM_PSHIFT_BITS)
#define PM_PSHIFT_MASK      (((1LL << PM_PSHIFT_BITS) - 1) << PM_PSHIFT_OFFSET)
#define PM_PSHIFT(x)        (((u64) (x) << PM_PSHIFT_OFFSET) & PM_PSHIFT_MASK)

#define PM_PFRAME_MASK      ((1LL << PM_PSHIFT_OFFSET) - 1)
#define PM_PFRAME(x)        ((x) & PM_PFRAME_MASK)

#define PM_PRESENT          PM_STATUS(4LL)
#define PM_SWAP             PM_STATUS(2LL)
#define PM_FILE             PM_STATUS(1LL)
//#define PM_NOT_PRESENT      PM_PSHIFT(PAGE_SHIFT)
//#define PM_END_OF_BUFFER    1

/*
 * data structures
 */

static int		opt_raw;	/* for kernel developers */
static int		opt_list;	/* list pages (in ranges) */
static int		opt_no_summary;	/* don't show summary */
static pid_t		opt_pid;	/* process to walk */

#define MAX_ADDR_RANGES	1024
static int		nr_addr_ranges;
static unsigned long	opt_offset[MAX_ADDR_RANGES];
static unsigned long	opt_size[MAX_ADDR_RANGES];

#define MAX_VMAS	10240
static int		nr_vmas;
static unsigned int	pg_flags[MAX_VMAS];
static unsigned long	pg_start[MAX_VMAS];
static unsigned long	pg_end[MAX_VMAS];
static unsigned long 	pg_ino[MAX_VMAS];

#define MAX_BIT_FILTERS	64
static int		nr_bit_filters;
static uint64_t		opt_mask[MAX_BIT_FILTERS];
static uint64_t		opt_bits[MAX_BIT_FILTERS];

static int		page_size;

static int		pagemap_fd;
static int		kpageflags_fd;

static int		opt_hwpoison;
static int		opt_unpoison;

static char		hwpoison_debug_fs[MAX_PATH+1];
static int		hwpoison_inject_fd;
static int		hwpoison_forget_fd;

#define HASH_SHIFT	13
#define HASH_SIZE	(1 << HASH_SHIFT)
#define HASH_MASK	(HASH_SIZE - 1)
#define HASH_KEY(flags)	(flags & HASH_MASK)

static unsigned long	total_pages;
static unsigned long	nr_pages[HASH_SIZE];
static uint64_t 	page_flags[HASH_SIZE];

/*
 * helper functions
 */

#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

#define min_t(type, x, y) ({			\
	type __min1 = (x);			\
	type __min2 = (y);			\
	__min1 < __min2 ? __min1 : __min2; })

#define max_t(type, x, y) ({			\
	type __max1 = (x);			\
	type __max2 = (y);			\
	__max1 > __max2 ? __max1 : __max2; })

static int checked_open(const char *pathname, int flags)
{
	int fd = open(pathname, flags);

	if (fd < 0) {
		perror(pathname);
		exit(EXIT_FAILURE);
	}

	return fd;
}

/*
 * pagemap/kpageflags routines
 */

static unsigned long do_u64_read(int fd, char *name,
				 uint64_t *buf,
				 unsigned long index,
				 unsigned long count)
{
	long bytes;

	if (index > ULONG_MAX / 8) {
		printf("index overflow: %lu\n", index);
		exit(1);
	}

	if (lseek(fd, index * 8, SEEK_SET) < 0) {
		perror(name);
		exit(EXIT_FAILURE);
	}

	bytes = read(fd, buf, count * 8);
	if (bytes < 0) {
		perror(name);
		exit(EXIT_FAILURE);
	}
	if (bytes % 8) {
		printf("partial read: %lu bytes\n", bytes);
		exit(1);
	}

	return bytes / 8;
}

static unsigned long pagemap_read(uint64_t *buf,
				  unsigned long index,
				  unsigned long pages)
{
	return do_u64_read(pagemap_fd, "/proc/pid/pagemap", buf, index, pages);
}
 
static unsigned long long parse_number(const char *str)
{
	unsigned long long n;

	n = strtoll(str, NULL, 0);

	if (n == 0 && str[0] != '0') {
		printf("invalid name or number: %s\n", str);
		exit(1);
	}

	return n;
}
 
static void parse_pid(const char *str)
{
	FILE *file;
	char buf[5000];

	opt_pid = parse_number(str);

	sprintf(buf, "/proc/%d/pagemap", opt_pid);
	pagemap_fd = checked_open(buf, O_RDONLY);

	sprintf(buf, "/proc/%d/maps", opt_pid);
	file = fopen(buf, "r");
	if (!file) {
		perror(buf);
		exit(EXIT_FAILURE);
	}

	while (fgets(buf, sizeof(buf), file) != NULL) {
		unsigned long vm_start;
		unsigned long vm_end;
		unsigned long long pgoff;
		int major, minor;
		char r, w, x, s;
		unsigned long ino;
		int n;

		n = sscanf(buf, "%lx-%lx %c%c%c%c %llx %x:%x %lu",
			   &vm_start,
			   &vm_end,
			   &r, &w, &x, &s,
			   &pgoff,
			   &major, &minor,
			   &ino);
		if (n < 10) {
			fprintf(stderr, "unexpected line: %s\n", buf);
			continue;
		}
		pg_start[nr_vmas] = vm_start / page_size;
		pg_end[nr_vmas] = vm_end / page_size;
		pg_flags[nr_vmas] = ((r == 'r') ? (1 << VMA_READ) : 0) |
				    ((w == 'w') ? (1 << VMA_WRITE) : 0) |
				    ((x == 'x') ? (1 << VMA_EXEC) : 0) |
				    ((s == 's') ? (1 << VMA_SHARE) : 0);
		pg_ino[nr_vmas] = ino;
		if (++nr_vmas >= MAX_VMAS) {
			fprintf(stderr, "too many VMAs\n");
			break;
		}
	}
	fclose(file);
}

static unsigned long pagemap_pfn(uint64_t val)
{
	unsigned long pfn;

	if (val & PM_PRESENT)
		pfn = PM_PFRAME(val);
	else
		pfn = 0;

	return pfn;
}

static unsigned long pagemap_swap(uint64_t val)
{
	unsigned long swap;
	
	if (val & PM_SWAP)
		swap = PM_PFRAME(val);
	else
		swap=0;
	
	return swap;
}

#define GET_PM_PSHIFT(x) (((unsigned long) (x) >> PM_PSHIFT_OFFSET) & ((1LL << PM_PSHIFT_BITS) - 1))

#define PRINT_READ(flags) ((flags & (1 << VMA_READ)) ? 'r' : '-')
#define PRINT_WRITE(flags) ((flags & (1 << VMA_WRITE)) ? 'w' : '-')
#define PRINT_EXEC(flags) ((flags & (1 << VMA_EXEC)) ? 'x' : '-')
#define PRINT_SHARE(flags) ((flags & (1 << VMA_SHARE)) ? 's' : 'p')
#define PRINT_FLAGS(flags) PRINT_READ(flags),PRINT_WRITE(flags),PRINT_EXEC(flags),PRINT_SHARE(flags)

unsigned long entries =0;

#define PAGEMAP_BATCH	(64 << 10)
static void walk_vma(unsigned long index, unsigned long count, unsigned int flags, unsigned long ino)
{
	uint64_t buf[PAGEMAP_BATCH];
	unsigned long batch;
	unsigned long pages;
	unsigned long pfn, swap, shift;
	unsigned long i;
	unsigned long buffer[3];

	while (count) {
		batch = min_t(unsigned long, count, PAGEMAP_BATCH);
		pages = pagemap_read(buf, index, batch);
		if (pages == 0)
			break;

		for (i = 0; i < pages; i++) {
                       if (buf[i] == 0x00) {
                                //printf("%08lx noPte noSwap noHuge\n", (index+i)*page_size);
                                continue;
                        }
		  
			shift = GET_PM_PSHIFT(buf[i]);
			if (shift != 12)
			  dprintf(STDERR_FILENO, "WARN this is a bigger page! %d bit of SHIFT\n", shift);
#if 0
			pfn = pagemap_pfn(buf[i]);
			if (pfn) {
			  // NOTE that kernel 3.2.14 don't support PM_FILE that tests on !PageAnon() therefore we cannot make this test later in tool.c (we should avoid using PM_FILE
/*			  if ( ((ino==0) && ((buf[i] & PM_FILE)!=0 )) ||
				((ino!=0) && ((buf[i] & PM_FILE) == 0)) )
			    printf("ERROR ino %ld buf %lx %lx (%lx)\n", ino, buf[i], (unsigned long)(buf[i] & PM_FILE), (unsigned long)PM_FILE ); */
			  
			  if (buf[i] & PM_FILE)
			    printf("%08lx pfn: %lx [%lx] %lx FILE %c%c%c%c\n", (index +i)*page_size, pfn, buf[i], shift, PRINT_FLAGS(flags));
			  else
			    printf("%08lx pfn: %lx [%lx] %lx %c%c%c%c\n", (index +i)*page_size, pfn, buf[i], shift, PRINT_FLAGS(flags));
			}
			else {
			  swap = pagemap_swap(buf[i]);
			  if (swap)
			    printf("%08lx swap: %lx [%lx] %lx %c%c%c%c\n", (index+i)*page_size, swap, buf[i], shift, PRINT_FLAGS(flags));
			  else
			    printf("%08lx none: %lx %lx %c%c%c%c\n", (index+i)*page_size, buf[i], shift, PRINT_FLAGS(flags));
			}
/*			if (pfn)
				walk_pfn(index + i, pfn, 1); */
#else
			//HERE I SHOULD JUST OUTPUT A TRI-PLET WITH
			// VADDR pagemapPFN FLAGS 
			buffer[0] = (index+i)*page_size;
			buffer[1] = buf[i];
			buffer[2] = (unsigned long) flags;
			{
			  ssize_t ret = write(STDOUT_FILENO, buffer, (3 *sizeof(unsigned long)));
			  if (ret < 0) { 
			    perror("ERROR while writing on stdout");
			    exit(1);
			  }
			}
			entries++;
#endif			
		}

		index += pages;
		count -= pages;
	}
}

static void walk_task(unsigned long index, unsigned long count, unsigned int flags, unsigned long ino)
{
	const unsigned long end = index + count;
	unsigned long start;
	int i = 0;

	while (index < end) {

		while (pg_end[i] <= index)
			if (++i >= nr_vmas)
				return;
		if (pg_start[i] >= end)
			return;

		start = max_t(unsigned long, pg_start[i], index);
		index = min_t(unsigned long, pg_end[i], end);

		assert(start < index);
		walk_vma(start, index - start, flags, ino);
	}
}

static void walk_pid()
{
	int i;
	for (i=0; i< nr_vmas; i++)
		walk_task(pg_start[i], (pg_end[i] - pg_start[i]), pg_flags[i], pg_ino[i]);
}

static void dump_pid()
{
	int i;
	for (i=0; i< nr_vmas; i++) {
		printf("vm:%d start:%lx(%lx) end:%lx(%lx) %c%c%c%c %d\n",
			i, 
			(pg_start[i] * page_size), pg_start[i], 
			(pg_start[i] * page_size), pg_end[i],
			(pg_flags[i] & (1 << VMA_READ)) ? 'r' : '-',
			(pg_flags[i] & (1 << VMA_WRITE)) ? 'w' : '-',
			(pg_flags[i] & (1 << VMA_EXEC)) ? 'x' : '-',
			(pg_flags[i] & (1 << VMA_SHARE)) ? 's' : 'p',
			(pg_end[i] - pg_start[i])
		);
		walk_task(pg_start[i], (pg_end[i] - pg_start[i]), pg_flags[i], pg_ino[i]);
	}
}

int main(int argc, char *argv[])
{
  dprintf(STDERR_FILENO, "This application merges /proc/pid/mmap and /proc/pid/pagemap\n" );
  dprintf(STDERR_FILENO, "in on standard out file with 8-byte triplets vaddr,pfn,vmaflags\n" );
  
  page_size= getpagesize(); // note that this is should be done better
  
  parse_pid(argv[1]);
  walk_pid();
/*  
  printf("%016lx\n%016lx\n%016lx\n%016lx PRESENT\n%016lx SWAP\n%016lx PRESENT and SWAP\n", PM_STATUS_MASK, PM_PSHIFT_MASK, PM_PFRAME_MASK, PM_PRESENT, PM_SWAP, (PM_PRESENT | PM_SWAP));
  dprintf(STDERR_FILENO, "entries: %ld\n", entries);
*/  

  return 0;
}
