 
/*
 * Antonio Barbalace, SSRG Virginia Tech, 2015
 */

#include <string.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/statfs.h>
#include <sys/stat.h>

#include "dproc.h"

static int page_size;

static long elems_array = 0;

/*
 * application's flags
 */
#define VMA_READ 0
#define VMA_WRITE 1
#define VMA_EXEC 2
#define VMA_SHARE 3

/*
 * pagemap kernel ABI bits
 */

#define PM_ENTRY_BYTES      sizeof(uint64_t)
#define PM_STATUS_BITS      3
#define PM_STATUS_OFFSET    (64 - PM_STATUS_BITS)
#define PM_STATUS_MASK      (((1LL << PM_STATUS_BITS) - 1) << PM_STATUS_OFFSET)
#define PM_STATUS(nr)       (((nr) << PM_STATUS_OFFSET) & PM_STATUS_MASK)

#define PM_PSHIFT_BITS      6
#define PM_PSHIFT_OFFSET    (PM_STATUS_OFFSET - PM_PSHIFT_BITS)
#define PM_PSHIFT_MASK      (((1LL << PM_PSHIFT_BITS) - 1) << PM_PSHIFT_OFFSET)
#define PM_PSHIFT(x)        (((u64) (x) << PM_PSHIFT_OFFSET) & PM_PSHIFT_MASK)

#define PM_PFRAME_MASK      ((1LL << PM_PSHIFT_OFFSET) - 1)
#define PM_PFRAME(x)        ((x) & PM_PFRAME_MASK)

#define PM_PRESENT          PM_STATUS(4LL)
#define PM_SWAP             PM_STATUS(2LL)
#define PM_FILE             PM_STATUS(1LL)
//#define PM_NOT_PRESENT      PM_PSHIFT(PAGE_SHIFT)
//#define PM_END_OF_BUFFER    1


static unsigned long pagemap_pfn(uint64_t val)
{
	unsigned long pfn;

	if (val & PM_PRESENT)
		pfn = PM_PFRAME(val);
	else
		pfn = 0;

	return pfn;
}

static unsigned long pagemap_swap(uint64_t val)
{
	unsigned long swap;
	
	if (val & PM_SWAP)
		swap = PM_PFRAME(val);
	else
		swap=0;
	
	return swap;
}

#define GET_PM_PSHIFT(x) (((unsigned long) (x) >> PM_PSHIFT_OFFSET) & ((1LL << PM_PSHIFT_BITS) - 1))

#define PRINT_READ(flags) ((flags & (1 << VMA_READ)) ? 'r' : '-')
#define PRINT_WRITE(flags) ((flags & (1 << VMA_WRITE)) ? 'w' : '-')
#define PRINT_EXEC(flags) ((flags & (1 << VMA_EXEC)) ? 'x' : '-')
#define PRINT_SHARE(flags) ((flags & (1 << VMA_SHARE)) ? 's' : 'p')
#define PRINT_FLAGS(flags) PRINT_READ(flags),PRINT_WRITE(flags),PRINT_EXEC(flags),PRINT_SHARE(flags)



unsigned long load_array(char * input_file, long elements)
{
  long l, bytes;
  unsigned long shift, pfn, swap;
  dproc_desc dproc;
  int pfd;
  
  pfd= open(input_file,O_RDONLY);
  if (pfd < 0) {
    perror(input_file);
    exit(1);
  }
  
  for (l=0; l<elements; l++) {
    bytes = read (pfd, (void *)&dproc, sizeof(dproc));
    if ( bytes < 0) {
      dprintf(STDERR_FILENO, "ERROR in reading the file %s\n", input_file);
      return -1;
    }
    if ( bytes % sizeof(dproc) ) {
      dprintf(STDERR_FILENO, "ERROR in the alignement, read returned %ld (%ld) for %s\n", bytes, sizeof(dproc), input_file);
      return -1;
    }
    
	shift = GET_PM_PSHIFT(dproc.info); // shift is page size NOTE
	if (shift != 12)
	  dprintf(STDERR_FILENO, "WARN this is a bigger page! %d bit of SHIFT\n", shift);

	pfn = pagemap_pfn(dproc.info);
	if (pfn) {
	  if (dproc.info & PM_FILE)
	    printf("%08lx pfn: %lx [%lx] %lx FILE %c%c%c%c\n", dproc.addr, pfn, dproc.info, shift, PRINT_FLAGS(dproc.flags));
	  else
	    printf("%08lx pfn: %lx [%lx] %lx ANON %c%c%c%c\n", dproc.addr, pfn, dproc.info, shift, PRINT_FLAGS(dproc.flags));
	}
	else {
	  swap = pagemap_swap(dproc.info);
	  if (swap)
	    printf("%08lx swap: %lx [%lx] %lx %c%c%c%c\n", dproc.addr, swap, dproc.info, shift, PRINT_FLAGS(dproc.flags));
	  else
	    printf("%08lx none: %lx %lx %c%c%c%c\n", dproc.addr, dproc.info, shift, PRINT_FLAGS(dproc.flags));
	}
   }
}


int main (int argc, char * argv []) 
{
  // check arguments, here load the paddr file and the pids mappings
  unsigned long l;
  char * input_dproc = "dproc.stdout";
  
  page_size= getpagesize(); // we are assuming, as documents online that the bock size is a page
  
  //allocate memory first
  struct stat st;
  if (stat (input_dproc, &st) != 0) {
    dprintf(STDERR_FILENO, "ERROR cannot determine the size of %s\n", input_dproc);
    exit(1);
  }
  elems_array = (st.st_size / sizeof(dproc_desc));
  
  printf("file '%s' size %ld elements %ld(%ld) (elem size %d)\n",
	input_dproc, st.st_size, elems_array,
	(st.st_size % sizeof(dproc_desc)), (int)sizeof(dproc_desc));

  //re-build the page struct array
/*  pages_array = malloc (elems_array * sizeof(struct page));
  if (pages_array == 0) {
    dprintf(STDERR_FILENO, "ERROR malloc failure for size of %ld\n", elems_array * sizeof(struct page));
    exit(1);
  } */
  load_array(input_dproc, elems_array); 
  
  return 0;
}